#include <linux/module.h>

#include <linux/kernel.h> /* printk() */
#include <linux/malloc.h> /* kmalloc() */
#include <linux/fs.h>     /* everything... */
#include <linux/errno.h>  /* error codes */
#include <linux/types.h>  /* size_t */
#include <linux/stddef.h>
#include <linux/proc_fs.h>
#include <linux/fcntl.h>        /* O_ACCMODE */
#include <linux/ioport.h>
#include <linux/pci.h>
#include <linux/sched.h>
#include <linux/tqueue.h>
#include <linux/interrupt.h>

#include <asm/uaccess.h>
#include <asm/system.h>   /* cli(), *_flags */
#include <asm/segment.h>  /* memcpy and such */
#include <asm/io.h>
/*
#include "svgalib_helper.h"
*/
int num_devices=0;

static int svgalib_helper_open( struct inode *inode, struct file * filp) {

	int minor = MINOR(inode->i_rdev);

	if(minor>num_devices) return -ENODEV;

	MOD_INC_USE_COUNT;

	return 0;
}

static int svgalib_helper_release( struct inode *inode, struct file *filp) {
	MOD_DEC_USE_COUNT;

	return 0;
}

static int svgalib_helper_mmap(struct file *filp, struct vm_area_struct *vma) {
	int start=vma->vm_start;
	int end=vma->vm_end;
	int minor = MINOR(filp->f_dentry->d_inode->i_rdev);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0)
	int ofs=vma->vm_pgoff*PAGE_SIZE;
#else
	int ofs=vma->vm_offset;
#endif
	unsigned int i, r, t, len, mem;
	struct pci_dev *dev = 0;

    	if(pci_present()) {
		i = minor + 2;
        	while(--i) dev=pci_find_class(PCI_CLASS_DISPLAY_VGA<<8,dev);
		for(i = 0; i < 6; i ++) {
                	pci_read_config_dword(dev,16+4*i,&r);
                	if(!r) continue;
                    	pci_write_config_dword(dev,16+4*i,0xffffffff);
                    	pci_read_config_dword(dev,16+4*i,&t);
                    	pci_write_config_dword(dev,16+4*i,r);
                    	len = ~(t&~0xf)+1;
			mem = r & PCI_BASE_ADDRESS_MEM_MASK;
			if(!len) {
				continue;
			}
			if((r & PCI_BASE_ADDRESS_SPACE) == 
					PCI_BASE_ADDRESS_SPACE_IO){
		       	       continue;
			}
			/* Do not allow memory maps in the lower 1MB */
			if(mem < 0x100000){
		       	       continue;	
			}
			if((ofs >= mem) && ((ofs + (end - start)) <= 
						(mem + len))){
				vma -> vm_flags |= VM_IO;
				if (remap_page_range(start, ofs, end-start, 
							vma-> vm_page_prot)) 
					return -EAGAIN;
				return 0;
			}
			/* We are only allowing access to the primary
			   aperture since the secondary is often used
			   for memory-mapped registers.
			 */
			break;
		}
	}
	return -EAGAIN;
}

struct file_operations svgalib_helper_fops = {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0)
owner:   THIS_MODULE,
#endif
mmap:    svgalib_helper_mmap,	/* mmap */
open:    svgalib_helper_open,	/* open */
release: svgalib_helper_release	/* release */
	};

int init_module(void)
{
    int result;

    /*
     * Register your major, and accept a dynamic number
     */
    result = register_chrdev(VGACLIENT_HELPER_MAJOR, "vgaclient_helper", &svgalib_helper_fops);
    if (result < 0) {
        printk(KERN_WARNING "vgaclient_helper: can't get major %d\n",VGACLIENT_HELPER_MAJOR);
        return result;
    }

    EXPORT_NO_SYMBOLS;

    return 0; /* succeed */
}

void cleanup_module(void)
{
}
